# Header for the simulation of the Riptide
import pyibex as py
from roblib import *
import math
import random


class Riptide:
    # block the of system
    # input {u: control signal}
    # output {y: measurement signal}
    def __init__(self, p_0, v_0, R_0, noise_d, noise_pz, noise_R):
        # initialise with the initial state p0, v0, R0, wr0
        self.p = p_0  # position
        self.v = v_0  # velocity
        self.R = R_0  # orientation matrix

        self.noise_pz = noise_pz
        self.noise_R = noise_R
        self.noise_d = noise_d  # distance noise with marks

    def sys_update(self, u, current=np.zeros((3, 1))):  # update the state of the system (Euler method)
        wr = self.v * _B @ u[1:4]
        self.p = self.p + dt * (self.R @ array([[self.v, 0, 0]]).T + current)
        self.v = self.v + dt * (p1 * u[0, 0] ** 2 - p2 * self.v * abs(self.v))
        self.R = self.R @ expw(dt * wr)

    def measure(self):  # return the output of the system with noise
        y_pz = self.measure_depth()
        y_R = self.measure_orientation()
        y_d = self.measure_distances()
        return y_pz, y_R, y_d

    def measure_depth(self):
        return self.p[2, 0] + rand(self.noise_pz)

    def measure_orientation(self):
        R_tilde = rand_box(self.noise_R).reshape((3, 3))
        return R_tilde @ self.R

    def measure_distances(self):
        y_d = np.zeros((_N, 1))
        for j in range(_N):
            y_d[j, 0] = np.linalg.norm(self.p - np.array([_Marks[j, :]]).T) + rand(self.noise_d)
        return y_d




class Int_Observer:
    # block the of Interval observer
    # input {y_pz, y_d: measurements}
    # output {p_hat: estimated position}
    def __init__(self, noise_d, noise_pz, noise_R, box_k_init, noise_vr):

        self.p_hat = py.IntervalVector(3)  # estimated position
        self.vr_hat = py.Interval(0)  # estimated velocity

        self.box_k = box_k_init  # initial guess for the iteration k
        self.pose_box_list = []  # list of the contracted sub boxes

        self.noise_d = noise_d  # distance noise with marks
        self.noise_pz = noise_pz
        self.noise_R = noise_R
        self.noise_R_box = np.array([noise_R.subvector(0, 2),
                                     noise_R.subvector(3, 5),
                                     noise_R.subvector(6, 8)])
        self.noise_vr = noise_vr  # interval of confidence for the speed estimation

        self.dp_box_old = py.IntervalVector(3)
        self.yd_old = -1 * np.ones((_N, 1))  # previous distance measurement
        self.y_pz_old = 0

        self.flag_circle_behavior = 0

    def estim_p(self, y_R, y_pz, y_d):

        # contract the position box
        stack = [self.box_k.copy()]
        while len(stack) < _nb_box:

            if len(stack) == 0:  # can occur with bad speed estimation
                print("warning: emtpy observer stack - reinitialisation")
                stack = [_box_k_init.copy()]  # reset interval

            # contract the last box of the stack
            p_box = self.contract_pos(stack.pop(), y_d, y_pz, _Mx, _My, _Mz)

            # (optional) if not empty, contract with previous measurement (if there is one)
            # if not p_box.is_empty() and self.yd_old[0,0]>=0:
            #     p_box = p_box - self.dp_box_old # go to time t-1
            #     p_box = self.contract_pos(p_box, self.yd_old, self.y_pz_old, _Mx, _My, _Mz)
            #     p_box = p_box + self.dp_box_old # go back to time t

            # if not empty, bisect the box and push back in stack
            if not p_box.is_empty():
                index = p_box.extr_diam_index(False)  # index of max diameter
                p_box_1, p_box_2 = p_box.bisect(index)
                stack = [p_box_1, p_box_2] + stack
        self.pose_box_list = stack

        # update the estimation of the position as the middle of the biggest box
        self.p_hat = np.array([max_box(self.pose_box_list).mid()]).T

        gbox = group_boxes(self.pose_box_list)
        if gbox.max_diam() > 15:
            self.flag_circle_behavior = 1
        elif gbox.max_diam() < 10:
            self.flag_circle_behavior = 0

        # update the initial box for the contraction of the next iteration
        dp = dt * (self.noise_R_box @ y_R) @ array([[self.vr_hat, 0, 0]]).T
        dp_box = py.IntervalVector([py.Interval(dp[0, 0]), py.Interval(dp[1, 0]), py.Interval(dp[2, 0])])
        box_list2 = []
        for j in range(len(self.pose_box_list)):
            box_list2.append(self.pose_box_list[j] + dp_box)
        self.box_k = group_boxes(box_list2)

        # save variables
        self.yd_old = y_d
        self.y_pz_old = y_pz
        self.dp_box_old = dp_box
        return

    def contract_pos(self, p_box, y_d, y_pz, Mx, My,
                     Mz):  # contract the p_box according to the measurement y_d and y_pz
        # Contract the estimation of the position xyz_hat_k knowing the measurements of time iteration k
        y_pz_int = py.Interval(y_pz + self.noise_pz)
        p_box[2] = cpz(p_box[2], y_pz_int)  # contract with y_pz
        for n in range(5):  # 5 times
            for i in range(len(y_d)):  # for every measurement
                # contract with y_d
                y_d_int = py.Interval(y_d[i, 0] + self.noise_d)
                p_box[0], p_box[1], p_box[2] = cmark(p_box[0], p_box[1], p_box[2], Mx[i].copy(),
                                                     My[i].copy(), Mz[i].copy(), y_d_int.copy())
        return p_box

    def estim_v(self, u0):
        # foward estimation of the velocity
        self.vr_hat = py.Interval(math.sqrt(p1 / p2) * u0 + self.noise_vr)


class EKF:
    # block of the state estimation (Extended Kalman Filter)
    # input {u: control signal, y: measurement signal}
    # output {px_hat, py_hat, v_hat}
    def __init__(self, P0, Pa, Pb, px_hat0, py_hat0, vr_hat0):
        # initialise the kalman filter
        self.px_hat = px_hat0
        self.py_hat = py_hat0
        self.vr_hat = vr_hat0

        self.gx = np.zeros((_N, 1))  # estimation of the distances from the markers

        self.Pa = Pa  # covariance of noise alpha
        self.Pb = Pb  # covariance of noise beta
        self.P = P0  # covariance matrix of the state

        self.Ak = np.eye(3)
        self.Ck = np.zeros((_N, 3))

    def update_Ak(self, R):  # update the evolution matrix Ak
        self.Ak = np.eye(3)
        self.Ak[0, 2] = dt * R[0, 0]
        self.Ak[1, 2] = dt * R[1, 0]
        self.Ak[2, 2] = 1 - 2 * p2 * abs(self.vr_hat)

    def update_Ck(self, pz):  # update the evolution matrix Ck
        self.Ck = np.zeros((_N, 3))
        for i in range(_N):
            p = np.array([[self.px_hat, self.py_hat, pz]]).T
            self.gx[i, 0] = np.linalg.norm(p - _Marks[[i], :].T)
            self.Ck[i] = np.array([(p[0, 0] - _Marks[i, 0]) / self.gx[i, 0],
                                   (p[1, 0] - _Marks[i, 1]) / self.gx[i, 0],
                                   0])

    def kalman_predict(self, u, R):  # prediction function of the kalman filter
        self.px_hat = self.px_hat + dt * R[0, 0] * self.vr_hat
        self.py_hat = self.py_hat + dt * R[1, 0] * self.vr_hat
        self.vr_hat = self.vr_hat + dt * (p1 * u[0, 0] ** 2 - p2 * self.vr_hat * abs(self.vr_hat))
        self.P = self.Ak @ self.P @ self.Ak.T + self.Pa

    def kalman_correct(self, y_d):
        # correction function of the kalman filter with the measurement vector y (for EKF only)
        zk = y_d - self.gx  # innovation
        Sk = self.Ck @ self.P @ self.Ck.T + self.Pb  # covariance of zk
        Kk = self.P @ self.Ck.T @ inv(Sk)  # kalman gain

        state = np.array([[self.px_hat, self.py_hat, self.vr_hat]]).T
        state = state + Kk @ zk
        self.px_hat, self.py_hat, self.vr_hat = state[0, 0], state[1, 0], state[2, 0]
        self.P = (eye(3) - Kk @ self.Ck) @ self.P

    def obs_out(self):  # return estimated state as a vector
        return self.px_hat, self.py_hat, self.vr_hat


class IEKF(EKF):
    # Extended Kalman Filter of the Interval Extended Kalman Filter
    # block of the state estimation
    # input {u: control signal, y: measurement signal}
    # output {px_hat, py_hat, v_hat}

    def update_Ck_IEKF(self, p_int):  # update the evolution matrix Ck at the point p_int
        self.Ck = np.zeros((_N, 3))
        for i in range(_N):
            self.gx[i, 0] = np.linalg.norm(p_int - _Marks[[i], :].T)
            self.Ck[i] = np.array([(p_int[0, 0] - _Marks[i, 0]) / self.gx[i, 0],
                                   (p_int[1, 0] - _Marks[i, 1]) / self.gx[i, 0],
                                   0])

    def kalman_correct_IEKF(self, y_d, p_int):  # correction function of the kalman filter with the measurement vector y
        zk = y_d - self.gx + self.Ck[:, 0:2] @ np.array(
            [[p_int[0, 0] - self.px_hat, p_int[1, 0] - self.py_hat]]).T  # innovation
        Sk = self.Ck @ self.P @ self.Ck.T + self.Pb  # covariance of zk
        Kk = self.P @ self.Ck.T @ inv(Sk)  # kalman gain

        state = np.array([[self.px_hat, self.py_hat, self.vr_hat]]).T
        state = state + Kk @ zk
        self.px_hat, self.py_hat, self.vr_hat = state[0, 0], state[1, 0], state[2, 0]
        self.P = (eye(3) - Kk @ self.Ck) @ self.P


class Controller:
    # controller block of the system
    # input {vd:velocity, Rk: line orientation, lk: line coordinates,
    #        p: estimated position of the riptide, R: measured orientation of the riptide,
    #        v: estimated velocity of the riptide}
    # output {u: control signal}
    def __init__(self):
        self.u = np.zeros((4, 1))
        self.Rd = np.eye(3)  # desired orientation matrix
        self.vd = 0

    def controller_update(self, vd, Rk, lk, p, R, v, flag_circle):
        if flag_circle:
            self.circle()
        else:
            self.vd = vd
            self.orient_towards_line(Rk, lk, p)
            self.reach_setpoint(R, v)

    def orient_towards_line(self, Rk, lk, p):  # from {Rk,lk,p} to {Rd}
        # rotation of angle θe along the axis ew to reduce the distance to the line e
        self.Rd = Rk
        xk = Rk[:, [0]]
        P, e = proj_on_line(p, xk, lk[:, [0]])
        if e > 0:
            θe = 0.5 * math.atan(e)  # gain 0.5 to reduce overshooting
            mp = normalize(P - p)
            ew = cross_col(xk, mp)
            self.Rd = expw(θe * ew) @ Rk

    def reach_setpoint(self, R, v):  # from (vd,Rd) to u
        self.u[0, 0] = math.sqrt(p2 / p1) * self.vd  # thruster

        # fins
        if abs(v) > 0.5:
            er = R.T @ logw(self.Rd @ R.T)
            if any(np.iscomplex(er)):  # if Rd @ x.R.T has no real logarithm, add small rotation to Rd
                er = R.T @ logw(eulermat(0, 0, 0.01) @ self.Rd @ R.T)
            self.u[1:4] = 1 / v * B_inv @ (er)

        um = max(abs(self.u[1:4]).T[0])
        if um > u_max:
            self.u[1:4] = (u_max / um) * self.u[1:4]  # saturation

    def circle(self):
        self.u[0, 0] = math.sqrt(p2 / p1) * 1
        er = 0.1 * np.array([[0, 0, -1]]).T
        self.u[1:4] = 1 / 1 * B_inv @ (er)
        um = max(abs(self.u[1:4]).T[0])
        if um > u_max:
            self.u[1:4] = (u_max / um) * self.u[1:4]  # saturation

    def set_setpoint(self,vd,Rd): # set the setpoint manually
        self.vd = vd
        self.Rd = Rd



class Record:  # made to memorise variable to then write them in the csv
    def __init__(self, t, p, vr, R, px_hat_kal, py_hat_kal, px_hat_int, py_hat_int, yz, u, P_p, pose_boxes, y_d, joint):
        self.t = t
        self.p = p
        self.vr = vr
        self.R = R
        self.px_hat_kal = px_hat_kal
        self.py_hat_kal = py_hat_kal
        self.px_hat_int = px_hat_int
        self.py_hat_int = py_hat_int
        self.yz = yz
        self.u = u
        self.P_p = P_p
        self.pose_boxes = pose_boxes
        self.y_d = y_d
        self.joint = joint


def cmark(x, y, z, mx, my, mz, d):  # contractor for the distance measurement of one marker
    Cd = py.CtcFwdBwd(fd, py.sqr(d))
    Bd = py.IntervalVector(6)
    Bd[0], Bd[1], Bd[2], Bd[3], Bd[4], Bd[5] = x, y, z, mx, my, mz
    Cd.contract(Bd)
    x, y, z = Bd[0], Bd[1], Bd[2]
    return x, y, z


def convert_param(L_dk, L_wk, p_0):  # compute the other parameters of the lines followed by the Riptide
    # input (line distances, line orientation vector, starting position)
    ak_list = [p_0]  # vertices
    Rk_list = []  # orientation matrices
    mk_list = []  # middles of the lines
    for j in range(len(L_dk)):
        Rk_list.append(expw(L_wk[:, [j]]))  # R(k) = exp(wk)
        ak_list.append(ak_list[j] + L_dk[j] * Rk_list[j][:, [0]])  # a(k+1) = a(k) + dk * xk
        mk_list.append(0.5 * (ak_list[j] + ak_list[j + 1]))  # m(k) = 0.5(a(k)+a(k+1))
    return np.concatenate(ak_list, 1), np.concatenate(mk_list, 1), Rk_list


def cpz(pz, y_pz):  # contractor for the depth measurement
    Cd = py.CtcFwdBwd(fpz, py.Interval(0))
    Bd = py.IntervalVector(2)
    Bd[0], Bd[1] = pz, y_pz
    Cd.contract(Bd)
    pz = Bd[0]
    return pz


def cross_col(a, b): return np.cross(a.T, b.T).T


def group_boxes(box_list):  # return the union of the boxes in box_list
    pose_box = box_list[0]
    for j in range(len(box_list) - 1):
        pose_box = pose_box | box_list[j + 1]
    return pose_box


def marks_interval(Marks, accuracy):  # return the intervals of the x y and z position of the marks
    # input (position of the markers, uncertainty of the position)
    Mx, My, Mz = [], [], []
    for N in range(len(Marks)):  # for every landmark
        Mx.append(py.Interval(Marks[N, 0] + accuracy))
        My.append(py.Interval(Marks[N, 1] + accuracy))
        Mz.append(py.Interval(Marks[N, 2] + accuracy))
    return Mx, My, Mz


def max_box(box_list):  # return the biggest box
    mbox = box_list[0]
    mmag = mbox.max_diam()
    for i in range(1, len(box_list)):
        if box_list[i].max_diam() > mmag:
            mbox = box_list[i]
    return mbox


def next_line(p, l):  # return true of the robot thinks it has passed the end of the line, estimating its position as p
    A, B = l[:, [0]], l[:, [1]]
    v1, v2 = p - B, A - B
    if v1.T[0].dot(v2.T[0]) < 0:
        return True
    return False


def normalize(v): return v / np.linalg.norm(v, 2)


def Pa():  # covariance matrix of the process noise (EKF)
    Pa = 0.01 * np.eye(3)
    return Pa


def Pb():  # covariance matrix of the measurement noise (EKF)
    Pb = 10 * eye(_N)
    return Pb


def proj_on_line(p, xk, A):  # project the estimated position p on the line of unit vector xk, passing by A
    P = A + (p - A).T[0].dot(xk.T[0]) * xk  # projection point
    e = np.linalg.norm(P - p, 2)  # distance to the line
    return P, e


def rand(In: py.Interval):  # take a random number in the interval
    return In.lb() + random.random() * (In.ub() - In.lb())


def rand_box(Box: py.IntervalVector):  # take a random vector in the box
    vec = []
    for Inter in Box:
        vec.append(rand(Inter))
    return np.array([vec]).T


# dynamic parameters
p1, p2, p3, p4, p5 = 0.1, 0.2, 1, 1, 1 #0.1, 0.2, 2, 0.5, 0.5
_B = array([[p3, 0, 0], [0, p4, 0], [0, 0, p5]]) @ array(
    [[-1, -1, -1], [0, math.sin(pi / 3), math.sin(-pi / 3)], [-1, math.cos(pi / 3), math.cos(pi / 3)]])
B_inv = inv(_B)
u_max = pi / 4  # fins' stop

# Mission parameters
_L_dk = [10, 20, 20, 10, 20, 10, 20, 10, 20]  # lengths of the lines
w1 = np.array([[0, 0, 0]]).T
w2 = pi / 4 * np.array([[0, 1, 0]]).T
w3 = -pi / 2 * np.array([[0, 0, 1]]).T
w4 = -pi * np.array([[0, 0, 1]]).T
w5 = pi / 2 * np.array([[0, 0, 1]]).T
w6 = -pi * np.array([[0, 0, 1]]).T
w7 = -pi / 2 * np.array([[0, 0, 1]]).T
w8 = -pi * np.array([[0, 0, 1]]).T
w9 = pi / 2 * np.array([[0, 0, 1]]).T
_L_wk = np.concatenate((w1, w2, w3, w4, w5, w6, w7, w8, w9), 1)  # list of vector of orientation
_vd = 1.  # desired speed

# initial state
_p_0 = np.array([[0, 0, -1]]).T  # x,y,z / real initial position
_p_0_hat = np.array([[0, 0, -1]]).T  # initial position guess
_v_0 = 0
_R_0 = expw(_L_wk[:, [0]])  # rotation matrix
_wr_0 = array([[0, 0, 0]]).T  # vx,vy,xz
_u_0 = np.zeros((4, 1))
β = 0  # angles for the blade

# noise
_noise_d = py.Interval(-0.5, 0.5)
_noise_pz = py.Interval(-0.1, 0.1)
na = py.Interval(-2 * pi / 360, 2 * pi / 360)
_noise_R = py.IntervalVector([py.cos(na), py.sin(na), py.sin(na),
                              py.sin(na), py.cos(na), py.sin(na),
                              py.sin(na), py.sin(na), py.cos(na)])
_noise_vr = py.Interval(-0.1, 0.1)  # for the interval analysis alone

# mission parameter bis
L_ak, L_mk, L_Rk = convert_param(_L_dk, _L_wk, _p_0_hat)
nl = 0  # number of the current line
_lk = L_ak[:, nl:nl + 2]  # current line lk
_Rk = L_Rk[nl]

# landmarks
_Marks = np.array([[0., -10., -1],
                   # [10., -15., -1],
                   [20., -10., -1]])  # [[lm_x,lm_y,lm_z]] true position of the landmarks
_N = len(_Marks)  # number of markers
_accuracy = py.Interval(-0.1, 0.1)  # accuracy of the landmark positioning
_Mx, _My, _Mz = marks_interval(_Marks, _accuracy)  # guess on the position of the landmarks

# contractor function
fd = py.Function("x", "y", "z", "mx", "my", "mz", "(x-mx)^2+(y-my)^2+(z-mz)^2")
fpz = py.Function("pz", "y_pz", 'pz-y_pz')

# Localisation variables
_P0 = 60 * eye(3)  # initial covariance matrix (col: px, py, vr)
_P0[2, 2] = 1
_Pa = Pa()
_Pb = Pb()
_nb_box = 10  # number of localisation boxes
_box_k_init = py.IntervalVector(
    [[-100, 100], [-100, 100], [-100, 0]])  # initial guess of the position given to the contractor

# current
_current = 0 * np.array([[-0.1, 0, 0]]).T

# simulation variables
tend = 100
dt = 0.1
L_t = arange(0, tend, dt)
